<?php

namespace Drupal\activitypub\Services;

use ActivityPhp\Server;
use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityFormBuilderInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Link;
use Drupal\Core\Pager\PagerManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\reader\ReaderBase;
use Symfony\Component\HttpFoundation\RequestStack;

class Reader extends ReaderBase {

  use StringTranslationTrait;

  /**
   * The ActivityPub Actor storage.
   *
   * @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface
   */
  protected $actorStorage;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The pager manager.
   *
   * @var \Drupal\Core\Pager\PagerManagerInterface
   */
  protected $pagerManager;

  /**
   * The request stack
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack $requestStack
   */
  protected $requestStack;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity form builder.
   *
   * @var \Drupal\Core\Entity\EntityFormBuilderInterface
   */
  protected $entityFormBuilder;

  /**
   * @var \Drupal\Core\Datetime\DateFormatInterface
   */
  protected $dateFormatter;

  /**
   * The ActivityPub Utility service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubUtilityInterface
   */
  protected $activityPubUtility;

  /**
   * The ActivityPub Media cache service.
   *
   * @var \Drupal\activitypub\Services\ActivityPubMediaCacheInterface
   */
  protected $activityPubMediaCache;

  /**
   * ActivityPubFormAlter constructor
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFormBuilderInterface $entity_form_builder
   *   The entity form builder.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Pager\PagerManagerInterface $pager_manager
   *   The pager manager.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   The date formatter.
   * @param \Drupal\activitypub\Services\ActivityPubUtilityInterface $activitypub_utility
   *   The ActivityPub utility service.
   * @param \Drupal\activitypub\Services\ActivityPubMediaCacheInterface $activitypub_media_cache
   *   The ActivityPub media cache service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityFormBuilderInterface $entity_form_builder, AccountInterface $current_user, RequestStack $request_stack, PagerManagerInterface $pager_manager, DateFormatterInterface $date_formatter, ActivityPubUtilityInterface $activitypub_utility, ActivityPubMediaCacheInterface $activitypub_media_cache) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFormBuilder = $entity_form_builder;
    $this->actorStorage = $entity_type_manager->getStorage('activitypub_actor');
    $this->currentUser = $current_user;
    $this->requestStack = $request_stack;
    $this->pagerManager = $pager_manager;
    $this->dateFormatter = $date_formatter;
    $this->activityPubUtility = $activitypub_utility;
    $this->activityPubMediaCache = $activitypub_media_cache;
  }

  /**
   * {@inheritdoc}
   */
  public function getChannels() {

    if (!$this->getActor()) {
      return [];
    }


    $conditions = [
      'uid' => $this->currentUser->id(),
      'status' => 1,
      'collection' => 'inbox',
      'is_read' => 0,
      'direct' => 0,
      'type' => ['Create', 'Like', 'Announce'],
    ];

    $following = $this->getActors(ActivityPubActivityInterface::FOLLOWING);
    if (!empty($following)) {
      $conditions['actor'] = $following;
    }

    $unread_home = $this->entityTypeManager->getStorage('activitypub_activity')->getActivityCount($conditions);
    $conditions['direct'] = 1;
    unset($conditions['actor']);
    $unread_direct = $this->entityTypeManager->getStorage('activitypub_activity')->getActivityCount($conditions);

    return [
      'label' => $this->t('Fediverse'),
      'channels' => [
        (object) [
          'uid' => 'home',
          'unread' => $unread_home,
          'name' => $this->t('Home')
        ],
        (object) [
          'uid' => 'direct',
          'unread' => $unread_direct,
          'name' => $this->t('Direct messages')
        ],
        (object) [
          'uid' => 'local',
          'unread' => 0,
          'name' => $this->t('Local timeline')
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getSourcesPage($op) {
    switch ($op) {

      case 'add':
        $build = $this->addSourcesButton($this->t('Back to follow collection'), 'activitypub');
        $activity = $this->entityTypeManager->getStorage('activitypub_activity')->create();
        $form = $this->entityFormBuilder->getForm($activity);
        $form['#action'] .= '?destination=/reader/sources/activitypub';
        $build['add'] = $form;
        break;

      case 'delete':
        $build = $this->addSourcesButton($this->t('Back to follow collection'), 'activitypub');
        $activity = $this->entityTypeManager->getStorage('activitypub_activity')->load($_GET['id']);
        if (!$activity->isPublished()) {
          $form = $this->entityFormBuilder->getForm($activity, 'delete');
          $form['actions']['cancel']['#access'] = FALSE;
          $form['#action'] .= '&destination=/reader/sources/activitypub';
          $build['delete'] = $form;
        }
        else {
          $build['delete'] = ['#markup' => '<div class="form-actions side-padding">' . $this->t(' This action cannot be undone.') . '</div>'];
          $build['actions'] = [
            '#prefix' => '<div class="form-actions side-padding">',
            '#suffix' => '</div>',
            'delete' => [
              '#type' => 'link',
              '#title' => 'delete',
              '#url' => $activity->toUrl('undo', ['query' => ['destination' => 'reader/sources/activitypub']]),
              '#attributes' => ['class' => ['button']],
            ],
          ];
        }
        $build['delete']['#prefix'] = $this->addSourcesConfirmDeleteText($activity->getObject());
        break;

      default:
        $build = $this->addSourcesButton($this->t('+ Follow'), 'activitypub', 'add');

        $rows = [];
        $following_properties = [
          'collection' => 'outbox',
          'type' => 'Follow',
        ];
        $following = $this->entityTypeManager->getStorage('activitypub_activity')->loadByProperties($following_properties);
        foreach ($following as $follow) {
          $row = [];
          $row[] = ['data' => ['#markup' => $follow->getObject() . (!$follow->isPublished() ? '<br /><div class="smaller-text">' . $this->t('Waiting for confirmation') . '</div>' : '')]];
          $row[] = Link::fromTextAndUrl($this->t('Unfollow'), Url::fromRoute('reader.sources', [
            'module' => 'activitypub',
            'op' => 'delete',
          ], ['query' => ['id' => $follow->id()]]))->toString();
          $rows[] = $row;
        }

        $build['table'] = [
          '#type' => 'table',
          '#rows' => $rows,
          '#empty' => $this->t('No followers available'),
          '#attributes' => ['class' => ['reader-content-table']],
        ];

        break;
    }

    $build['#title'] = $this->t('Manage Fediverse Followers');
    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function getTimelineActions($id) {

    if (!$this->getActor()) {
      return [];
    }

    $status_title = isset($_SESSION['activitypub_status']) ? $this->t('View all items') : $this->t('View unread items');

    return [
      ['action' => 'mark-read', 'title' => $this->t('Mark read')],
      ['action' => 'status', 'title' => $status_title]
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function doTimelineAction($action, $id) {
    if ($this->getActor()) {
      if ($action == 'mark-read') {
        $this->entityTypeManager->getStorage('activitypub_activity')->changeReadStatus(1, $id);
      }
      if ($action == 'status') {
        if (isset($_SESSION['activitypub_status'])) {
          unset($_SESSION['activitypub_status']);
        }
        else {
          $_SESSION['activitypub_status'] = TRUE;
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getPostActions($id, $item) {

    if (!$this->getActor()) {
      return [];
    }

    $action = 'mark-unread';
    $title = $this->t('Mark unread');
    if (isset($item->_is_read) && !$item->_is_read) {
      $action = 'mark-read';
      $title = $this->t('Mark read');
    }

    return [[
      'action' => $action,
      'title' => $title,
    ]];
  }

  /**
   * {@inheritdoc}
   */
  public function doPostAction($action, $id, $items) {
    if ($action == 'mark-read' || $action == 'mark-unread') {
      $status = $action == 'mark-read' ? 1 : 0;
      $this->entityTypeManager->getStorage('activitypub_activity')->changeReadStatus($status, $id, $items);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getTimeline($id, $search = NULL) {

    if (!$this->getActor()) {
      return ['items' => []];
    }

    return $this->getResponse($id, $search);
  }

  /**
   * Get the actor for the current user.
   *
   * @return bool|NULL|\Drupal\activitypub\Entity\ActivitypubActorInterface
   */
  protected function getActor() {
    if ($this->currentUser->hasPermission('access reader')) {
      return $this->actorStorage->loadActorByEntityIdAndType($this->currentUser->id(), 'person');
    }

    return FALSE;
  }

  /**
   * Get actors.
   *
   * @param $type
   *   The type of followers.
   *
   * @return array
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getActors($type) {
    $items = [];

    $actor = $this->getActor();
    $conditions = ['type' => 'Follow', 'status' => 1];
    $url = Url::fromRoute('activitypub.user.self', ['user' => $actor->getOwnerId(), 'activitypub_actor' => $actor->getName()], ['absolute' => TRUE])->toString();
    if ($type == ActivityPubActivityInterface::FOLLOWERS) {
      /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface[] $activities */
      $conditions['object'] = $url;
      $conditions['collection'] = ActivityPubActivityInterface::INBOX;
    }
    else {
      $conditions['actor'] = $url;
      $conditions['collection'] = ActivityPubActivityInterface::OUTBOX;
    }
    /** @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorageInterface $storage */
    $storage = $this->entityTypeManager->getStorage('activitypub_activity');
    $records = $storage->getActivityRecords($conditions);
    foreach ($records as $record) {
      $items[] = $type == ActivityPubActivityInterface::FOLLOWERS ? $record->actor : $record->object;
    }

    return $items;
  }

  /**
   * Get items for a timeline.
   *
   * @param $id
   *   The channel id
   * @param $search
   *   The search param.
   *
   * @return array
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getResponse($id, $search) {

    $items = [];
    $paging = [];

    $limit = Settings::get('reader_activitypub_timeline_limit', 20);

    // Get page.
    $page = $this->requestStack->getCurrentRequest()->get('page', 0);
    $server = $this->activityPubUtility->getServer();

    /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface[] $activities */
    $storage = $this->entityTypeManager->getStorage('activitypub_activity');
    $query = $storage->getQuery()
      ->condition('uid', $this->currentUser->id())
      ->condition('status', 1)
      ->condition('collection', 'inbox')
      ->condition('type', $this->activityPubUtility->getTimelineTypes(), 'IN')
      ->pager($limit)
      ->sort('id', 'DESC');
    if (isset($_SESSION['activitypub_status'])) {
      $query->condition('is_read', 0);
    }

    switch ($id) {
      case 'home':
        $query->condition('direct', 0);
        $following = $this->getActors(ActivityPubActivityInterface::FOLLOWING);
        if (!empty($following)) {
          $query->condition('actor', $following, 'IN');
        }
        break;
      case 'direct':
        $query->condition('direct', 1);
        break;
      default:
        $query->condition('direct', 0);
        break;
    }

    $ids = $query->execute();
    $activities = $storage->loadMultiple($ids);
    foreach ($activities as $activity) {
      if ($item = $this->buildMicrosubItem($activity, $server, $id)) {
        $items[] = $item;
      }
    }

    $pager = $this->pagerManager->getPager();
    $pager_total = $pager->getTotalPages();
    $page++;
    if ((isset($pager_total) && is_numeric($pager_total) && $pager_total > $page) || $page > 0) {
      $paging = ['after' => $page];
    }

    $response = ['items' => $items];
    if (!empty($paging)) {
      $response['paging'] = (object) $paging;
    }

    return $response;
  }

  /**
   * Build a Microsub item from an activity payload.
   *
   * @param \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity
   * @param \ActivityPhp\Server $server
   * @param string $id
   *
   * @return \stdClass $item
   */
  protected function buildMicrosubItem(ActivityPubActivityInterface $activity, Server $server, string $id) {

    $payload = json_decode($activity->getPayLoad(), TRUE);
    $context = json_decode($activity->getContext(), TRUE);

    $item = [];
    $item['_id'] = $activity->id();
    $item['_is_read'] = !($id == 'home' || $id == 'direct') || $activity->isRead();
    $item['type'] = 'entry';
    if ($activity->isDirectMessage()) {
      $item['url'] = !empty($payload['object']['id']) ? $payload['object']['id'] : str_replace('/activity', '', $activity->getExternalId());
    }
    else {
      $item['url'] = $activity->getObject();
    }
    $item['published'] = $this->dateFormatter->format($activity->getCreatedTime(), 'custom','Y-m-dTH:i:s');

    // Content and response type.
    if ($activity->getType() == 'Like') {
      $item['like-of'] = [$activity->getObject()];
    }
    elseif ($activity->getType() == 'Announce') {
      $item['repost-of'] = [$activity->getObject()];
    }
    elseif ($activity->getType() == 'Create') {
      if (!empty($payload['object']) && is_array($payload['object'])) {
        if (!empty($payload['object']['inReplyTo'])) {
          $item['in-reply-to'] = [$payload['object']['inReplyTo']];
        }

        if (!empty($payload['object']['name'])) {
          $item['name'] = $payload['object']['name'];
        }

        if (!empty($payload['object']['content'])) {
          $item['content'] = (object) ['html' => $payload['object']['content']];
        }

        if (!empty($payload['object']['attachment'])) {
          $this->handleAttachments($payload['object']['attachment'], $item);
        }

      }
      else {
        if (!empty($payload['name'])) {
          $item['name'] = $payload['name'];
        }
        else {
          // currently treat as repost
          $item['repost-of'] = [$activity->getObject()];
        }
      }
    }

    // Context.
    if (!empty($context) && (isset($item['repost-of']) || isset($item['like-of']) || isset($item['in-reply-to']))) {
      $url = NULL;
      if (isset($item['repost-of'])) {
        $url = $item['repost-of'][0];
      }
      elseif (isset($item['like-of'])) {
        $url = $item['like-of'][0];
      }
      elseif (isset($item['in-reply-to'])) {
        $url = $item['in-reply-to'][0];
      }

      if (isset($url) && isset($context['id']) && $context['id'] == $url) {
        $ref = [];
        if (!empty($context['name'])) {
          $ref['name'] = $context['name'];
        }
        if (!empty($context['content'])) {
          $ref['content'] = (object) ['html' => $context['content']];
        }
        if (!empty($context['summary'])) {
          $ref['summary'] = $context['summary'];
        }
        if (!empty($context['attachment'])) {
          $this->handleAttachments($context['attachment'], $ref);
        }

        $item['refs'] = new \stdClass();
        $item['refs']->{$url} = (object) $ref;
      }
    }

    // Author information.
    $name = $photo = '';
    $target_actor = $server->actor($activity->getActor());
    if ($target_actor) {
      $name = $target_actor->get('name');
      $icon = $target_actor->get('icon');
      if (!empty($icon)) {
        $photo = $this->activityPubMediaCache->applyImageCache($icon->get('url'));
      }
    }

    $item['author'] = (object) ['url' => $activity->getActor(), 'name' => $name, 'photo' => $photo];

    return (object) $item;
  }

  /**
   * Handle attachments.
   *
   * @param $attachments
   * @param $o
   */
  protected function handleAttachments($attachments, &$o) {
    foreach ($attachments as $attachment) {
      if (isset($attachment['mediaType'])) {
        switch ($attachment['mediaType']) {
          case 'image/jpg':
          case 'image/png':
          case 'image/jpeg':
          case 'image/gif':
            if (!isset($o['photo'])) {
              $o['photo'] = [];
            }
            $o['photo'][] = $this->activityPubMediaCache->applyImageCache($attachment['url'], 'cache_attachment');
            break;
          case 'video/mp4':
            if (!isset($o['video'])) {
              $o['video'] = [];
            }
            $o['video'][] = $attachment['url'];
            break;
          case 'audio/mp3':
            if (!isset($o['audio'])) {
              $o['audio'] = [];
            }
            $o['audio'][] = $attachment['url'];
            break;
        }
      }
    }
  }

}
